# -*- Mode: Python -*-
cimport cython

import numpy as np
cimport numpy as np

from sfepy.discrete.common.extmods.types cimport int32, uint32, float64

from sfepy.discrete.common.extmods._fmfield cimport (FMField,
                                                     array2fmfield4,
                                                     array2fmfield3,
                                                     array2fmfield2,
                                                     array2fmfield1,
                                                     array2pint2,
                                                     array2pint1,
                                                     array2puint1,
                                                     fmf_pretend_nc,
                                                     fmf_alloc,
                                                     fmf_free)

cdef extern from 'common.h':
    void *pyalloc(size_t size)
    void pyfree(void *pp)

cdef extern from 'nurbs.h':
    ctypedef struct NURBSContext:
        int32 (*get_xi_dist)(float64 *pdist, FMField *xi,
                             FMField *point, FMField *e_coors,
                             void *_ctx)
        int32 (*eval_basis)(FMField *out, FMField *coors, int32 diff,
                            void *_ctx)
        int32 iel # >= 0.
        int32 is_dx # 1 => apply reference mapping to gradient.
        FMField e_coors_max[1] # Buffer for coordinates of element nodes.

        FMField control_points[1]
        FMField weights[1]
        int32 *degrees
        int32 dim
        FMField cs[3]
        int32 *conn
        int32 n_cell
        int32 n_efun

        FMField bf[1]
        FMField bfg[1]

        FMField R[1]
        FMField dR_dxi[1]
        FMField dR_dx[1]

        FMField B[3]
        FMField dB_dxi[3]
        FMField N[3]
        FMField dN_dxi[3]

        int32 reuse

        int32 has_bernstein
        int32 is_nurbs

        int32 i_max
        float64 newton_eps

    cdef void _ravel_multi_index \
         'ravel_multi_index'(uint32 *index, uint32 *indices,
                             uint32 *shape, uint32 num)
    cdef void _unravel_index \
         'unravel_index'(uint32 *indices, uint32 index,
                         uint32 *shape, uint32 num)

    void _print_context_nurbs \
         'print_context_nurbs'(NURBSContext *ctx)

    int32 _get_xi_dist \
          'get_xi_dist'(float64 *pdist, FMField *xi,
                        FMField *point, FMField *e_coors,
                        void *_ctx)

    int32 _eval_basis_nurbs \
          'eval_basis_nurbs'(FMField *out, FMField *coors, int32 diff,
                             void *_ctx)

    cdef int32 _eval_bernstein_basis \
         'eval_bernstein_basis'(FMField *funs, FMField *ders,
                                float64 x, uint32 degree)
    cdef int32 _eval_bspline_basis_tp \
         'eval_bspline_basis_tp'(FMField *R, FMField *dR_dx, FMField *det,
                                 FMField *dR_dxi,
                                 FMField *dx_dxi, FMField *dxi_dx,
                                 FMField *B, FMField *dB_dxi,
                                 FMField *N, FMField *dN_dxi,
                                 FMField *qp, uint32 ie,
                                 FMField *control_points,
                                 int32 *degrees, int32 dim,
                                 FMField *cs,
                                 int32 *conn, int32 n_el, int32 n_ep,
                                 int32 has_bernstein, int32 is_dx)
    cdef int32 _eval_nurbs_basis_tp \
         'eval_nurbs_basis_tp'(FMField *R, FMField *dR_dx, FMField *det,
                               FMField *dR_dxi,
                               FMField *dx_dxi, FMField *dxi_dx,
                               FMField *B, FMField *dB_dxi,
                               FMField *N, FMField *dN_dxi,
                               FMField *qp, uint32 ie, FMField *control_points,
                               FMField *weights, int32 *degrees, int32 dim,
                               FMField *cs,
                               int32 *conn, int32 n_el, int32 n_ep,
                               int32 has_bernstein, int32 is_dx)

cdef class CNURBSContext:

    cdef NURBSContext *ctx

    # Store arrays to prevent their deallocation in Python.
    cdef readonly np.ndarray e_coors_max # Auxiliary buffer.

    cdef readonly np.ndarray bf # Auxiliary buffer.
    cdef readonly np.ndarray bfg # Auxiliary buffer.

    cdef readonly np.ndarray R # Auxiliary buffer.
    cdef readonly np.ndarray dR_dxi # Auxiliary buffer.
    cdef readonly np.ndarray dR_dx # Auxiliary buffer.

    cdef readonly np.ndarray bufBN # Auxiliary buffer.

    property iel:

        def __get__(self):
            return self.ctx.iel

        def __set__(self, int32 iel):
            assert iel < self.ctx.n_cell
            self.ctx.iel = iel

    def __cinit__(self,
                  np.ndarray[float64, mode='c', ndim=2] control_points not None,
                  np.ndarray[float64, mode='c', ndim=1] weights not None,
                  np.ndarray[int32, mode='c', ndim=1] degrees not None,
                  cs not None,
                  np.ndarray[int32, mode='c', ndim=2] conn not None,
                  int32 i_max=100,
                  float64 newton_eps=1e-8):
        cdef NURBSContext *ctx
        cdef int32 ii, num
        cdef float64 *buf
        cdef np.ndarray[float64, mode='c', ndim=2] _e_coors_max
        cdef np.ndarray[float64, mode='c', ndim=1] _bf
        cdef np.ndarray[float64, mode='c', ndim=2] _bfg
        cdef np.ndarray[float64, mode='c', ndim=1] _R
        cdef np.ndarray[float64, mode='c', ndim=2] _dR_dxi
        cdef np.ndarray[float64, mode='c', ndim=2] _dR_dx
        cdef np.ndarray[float64, mode='c', ndim=1] _bufBN

        ctx = self.ctx = <NURBSContext *> pyalloc(sizeof(NURBSContext))

        if ctx is NULL:
            raise MemoryError()

        ctx.get_xi_dist = &_get_xi_dist
        ctx.eval_basis = &_eval_basis_nurbs
        ctx.iel = 0
        ctx.is_dx = 1

        array2fmfield2(ctx.control_points, control_points)
        array2fmfield1(ctx.weights, weights)

        array2pint1(&ctx.degrees, &ctx.dim, degrees)

        for ii in range(ctx.dim):
            array2fmfield4(ctx.cs + ii, cs[ii])

        array2pint2(&ctx.conn, &ctx.n_cell, &ctx.n_efun, conn)

        _e_coors_max = self.e_coors_max = np.zeros((ctx.n_efun, ctx.dim),
                                                   dtype=np.float64)
        array2fmfield2(ctx.e_coors_max, _e_coors_max)

        _bf = self.bf = np.zeros((ctx.n_efun,), dtype=np.float64)
        array2fmfield1(ctx.bf, _bf)

        _bfg = self.bfg = np.zeros((ctx.dim, ctx.n_efun),
                                   dtype=np.float64)
        array2fmfield2(ctx.bfg, _bfg)

        _R = self.R = np.zeros((ctx.n_efun,), dtype=np.float64)
        array2fmfield1(ctx.R, _R)

        _dR_dxi = self.dR_dxi = np.zeros((ctx.dim, ctx.n_efun),
                                         dtype=np.float64)
        array2fmfield2(ctx.dR_dxi, _dR_dxi)

        _dR_dx = self.dR_dx = np.zeros((ctx.dim, ctx.n_efun),
                                       dtype=np.float64)
        array2fmfield2(ctx.dR_dx, _dR_dx)

        n_efuns = degrees + 1

        _bufBN = self.bufBN = np.zeros((4 * n_efuns.sum(),), dtype=np.float64)
        buf = &_bufBN[0]
        for ii in range(0, ctx.dim):
            num = n_efuns[ii]
            fmf_pretend_nc(ctx.B + ii, 1, 1, num, 1, buf)
            buf += num
            fmf_pretend_nc(ctx.dB_dxi + ii, 1, 1, num, 1, buf)
            buf += num
            fmf_pretend_nc(ctx.N + ii, 1, 1, num, 1, buf)
            buf += num
            fmf_pretend_nc(ctx.dN_dxi + ii, 1, 1, num, 1, buf)
            buf += num

        ctx.reuse = 0

        ctx.is_nurbs = is_nurbs(weights)

        ctx.i_max = i_max
        ctx.newton_eps = newton_eps

    def __dealloc__(self):
        pyfree(self.ctx)

    def __str__(self):
        return 'CNURBSContext'

    def cprint(self):
        _print_context_nurbs(self.ctx)

    def evaluate(self, np.ndarray[float64, mode='c', ndim=2] coors not None,
                 int32 diff=False, **kwargs):
        cdef int32 n_coor = coors.shape[0]
        cdef int32 n_efun = self.ctx.n_efun
        cdef int32 dim = coors.shape[1]
        cdef int32 bdim
        cdef FMField[1] _out, _coors

        ctx = self.ctx

        if diff:
            bdim = dim

        else:
            bdim = 1

        cdef np.ndarray[float64, ndim=3] out = np.zeros((n_coor, bdim, n_efun),
                                                        dtype=np.float64)

        array2fmfield3(_out, out)
        array2fmfield2(_coors, coors)

        self.ctx.eval_basis(_out, _coors, diff, ctx)

        return out

def is_nurbs(np.ndarray[float64, mode='c', ndim=1] weights not None):
    """
    Return True if some weights are not one.
    """
    return np.any(weights != 1.0)

def eval_bernstein_basis(np.ndarray funs not None,
                         np.ndarray ders not None,
                         float64 x,
                         uint32 degree):
    cdef int32 ret
    cdef FMField[1] _funs, _ders

    array2fmfield1(_funs, funs)
    array2fmfield1(_ders, ders)

    ret = _eval_bernstein_basis(_funs, _ders, x, degree)
    return ret

def eval_mapping_data_in_qp(np.ndarray[float64, mode='c', ndim=2] qps not None,
                            np.ndarray[float64, mode='c', ndim=2]
                            control_points not None,
                            np.ndarray[float64, mode='c', ndim=1]
                            weights not None,
                            np.ndarray[int32, mode='c', ndim=1]
                            degrees not None,
                            cs not None,
                            np.ndarray[int32, mode='c', ndim=2] conn not None,
                            np.ndarray[uint32, mode='c', ndim=1] cells=None):
    """
    Evaluate data required for the isogeometric domain reference mapping in the
    given quadrature points. The quadrature points are the same for all Bezier
    elements and should correspond to the Bernstein basis degree.

    Parameters
    ----------
    qps : array
        The quadrature points coordinates with components in [0, 1] reference
        element domain.
    control_points : array
        The NURBS control points.
    weights : array
        The NURBS weights.
    degrees : sequence of ints or int
        The basis degrees in each parametric dimension.
    cs : list of lists of 2D arrays
        The element extraction operators in each parametric dimension.
    conn : array
        The connectivity of the global NURBS basis.
    cells : array, optional
        If given, use only the given Bezier elements.

    Returns
    -------
    bfs : array
        The NURBS shape functions in the physical quadrature points of all
        elements.
    bfgs : array
        The NURBS shape functions derivatives w.r.t. the physical coordinates
        in the physical quadrature points of all elements.
    dets : array
        The Jacobians of the mapping to the unit reference element in the
        physical quadrature points of all elements.
    """
    cdef uint32 ii, ie, n_qp, n_efun, nf
    cdef int32 n_el, n_ep, dim, aux
    cdef uint32 *_cells
    cdef (int32 *) _degrees, _conn
    cdef FMField[1] _bf, _bfg, _det
    cdef FMField[1] _bfg_dxi, _dx_dxi, _dxi_dx
    cdef FMField[1] _qp, _control_points, _weights
    cdef FMField _cs[3]
    cdef FMField[3] _B, _dB_dxi, _N, _dN_dxi
    cdef np.ndarray[float64, mode='c', ndim=4] bfs, bfgs, dets

    if cells is None:
        cells = np.arange(conn.shape[0], dtype=np.uint32)

    degrees = np.asarray(degrees, dtype=np.int32)

    n_el = len(cells)
    n_qp = qps.shape[0]
    dim = control_points.shape[1]
    n_efuns = degrees + 1
    n_efun = np.prod(n_efuns)

    # Output Jacobians.
    dets = np.empty((n_el, n_qp, 1, 1), dtype=np.float64)

    # Output shape functions.
    bfs = np.empty((n_el, n_qp, 1, n_efun), dtype=np.float64)

    # Output gradients of shape functions.
    bfgs = np.empty((n_el, n_qp, dim, n_efun), dtype=np.float64)

    # Setup C termporary arrays.
    bfg_dxi = np.empty((1, 1, dim, n_efun), dtype=np.float64)
    dx_dxi = np.empty((1, 1, dim, dim), dtype=np.float64)
    dxi_dx = np.empty((1, 1, dim, dim), dtype=np.float64)

    for ii in range(0, <uint32>dim):
        fmf_alloc(_B + ii, n_qp, 1, n_efuns[ii], 1)
        fmf_alloc(_dB_dxi + ii, n_qp, 1, n_efuns[ii], 1)
        fmf_alloc(_N + ii, 1, 1, n_efuns[ii], 1)
        fmf_alloc(_dN_dxi + ii, 1, 1, n_efuns[ii], 1)

    # Assign to C structures.
    array2fmfield4(_bfg_dxi, bfg_dxi)
    array2fmfield4(_dx_dxi, dx_dxi)
    array2fmfield4(_dxi_dx, dxi_dx)
    array2fmfield2(_control_points, control_points)
    array2fmfield1(_weights, weights)
    for ii in range(<uint32>dim):
        array2fmfield4(_cs + ii, cs[ii])

    array2pint1(&_degrees, &dim, degrees)
    array2pint2(&_conn, &aux, &n_ep, conn)

    _bf.offset = _bfg.offset = _det.offset = _qp.offset = -1
    _bf.nAlloc = _bfg.nAlloc = _det.nAlloc = _qp.nAlloc = -1
    _bf.nCell = _bfg.nCell = _det.nCell = _qp.nCell = 1
    _bf.nLev = _bfg.nLev = _det.nLev = _qp.nLev = 1
    _bf.nRow = _det.nRow = _qp.nRow = 1
    _bfg.nRow = dim
    _bf.nCol = _bfg.nCol = n_efun
    _det.nCol = 1
    _qp.nCol = dim

    _bf.val = _bf.val0 = &bfs[0, 0, 0, 0]
    _bfg.val = _bfg.val0 = &bfgs[0, 0, 0, 0]
    _det.val = _det.val0 = &dets[0, 0, 0, 0]
    _qp.val = _qp.val0 = &qps[0, 0]

    # Pre-compute 1D Bernstein basis B, dB/dxi.
    for iqp in range(0, n_qp):
        for ii in range(0, <uint32>dim):
            nf = n_efuns[ii]
            _eval_bernstein_basis(_B + ii, _dB_dxi + ii,
                                  _qp.val[ii], _degrees[ii])
            (_B + ii).val += nf
            (_dB_dxi + ii).val += nf
        _qp.val += dim

    if is_nurbs(weights):
        # Loop over elements.
        _cells = &cells[0]
        for iseq in range(0, n_el):
            ie = _cells[iseq]

            _qp.val = _qp.val0
            for ii in range(0, <uint32>dim):
                (_B + ii).val = (_B + ii).val0
                (_dB_dxi + ii).val = (_dB_dxi + ii).val0

            # Loop over quadrature points.
            for iqp in range(0, n_qp):
                _eval_nurbs_basis_tp(_bf, _bfg, _det,
                                     _bfg_dxi,
                                     _dx_dxi, _dxi_dx,
                                     _B, _dB_dxi, _N, _dN_dxi,
                                     _qp, ie,
                                     _control_points, _weights,
                                     _degrees, dim, _cs, _conn, n_el, n_ep,
                                     1, 1)
                _bf.val += n_efun
                _bfg.val += dim * n_efun
                _det.val += 1
                _qp.val += dim
                for ii in range(0, <uint32>dim):
                    nf = n_efuns[ii]
                    (_B + ii).val += nf
                    (_dB_dxi + ii).val += nf

    else:
        # Loop over elements.
        _cells = &cells[0]
        for iseq in range(0, n_el):
            ie = _cells[iseq]

            _qp.val = _qp.val0
            for ii in range(0, <uint32>dim):
                (_B + ii).val = (_B + ii).val0
                (_dB_dxi + ii).val = (_dB_dxi + ii).val0

            # Loop over quadrature points.
            for iqp in range(0, n_qp):
                _eval_bspline_basis_tp(_bf, _bfg, _det,
                                       _bfg_dxi,
                                       _dx_dxi, _dxi_dx,
                                       _B, _dB_dxi, _N, _dN_dxi,
                                       _qp, ie,
                                       _control_points,
                                       _degrees, dim, _cs, _conn, n_el, n_ep,
                                       1, 1)
                _bf.val += n_efun
                _bfg.val += dim * n_efun
                _det.val += 1
                _qp.val += dim
                for ii in range(0, <uint32>dim):
                    nf = n_efuns[ii]
                    (_B + ii).val += nf
                    (_dB_dxi + ii).val += nf

    for ii in range(0, <uint32>dim):
        fmf_free(_B + ii)
        fmf_free(_dB_dxi + ii)
        fmf_free(_N + ii)
        fmf_free(_dN_dxi + ii)

    return bfs, bfgs, dets

def eval_variable_in_qp(np.ndarray[float64, mode='c', ndim=2] variable not None,
                        np.ndarray[float64, mode='c', ndim=2] qps not None,
                        np.ndarray[float64, mode='c', ndim=2]
                        control_points not None,
                        np.ndarray[float64, mode='c', ndim=1] weights not None,
                        np.ndarray[int32, mode='c', ndim=1] degrees not None,
                        cs not None,
                        np.ndarray[int32, mode='c', ndim=2] conn not None,
                        np.ndarray[uint32, mode='c', ndim=1] cells=None):
    """
    Evaluate a field variable in the given quadrature points. The quadrature
    points are the same for all Bezier elements and should correspond to the
    Bernstein basis degree. The field variable is defined by its DOFs - the
    coefficients of the NURBS basis.

    Parameters
    ----------
    variable : array
        The DOF values of the variable with n_c components, shape (:, n_c).
    qps : array
        The quadrature points coordinates with components in [0, 1] reference
        element domain.
    control_points : array
        The NURBS control points.
    weights : array
        The NURBS weights.
    degrees : sequence of ints or int
        The basis degrees in each parametric dimension.
    cs : list of lists of 2D arrays
        The element extraction operators in each parametric dimension.
    conn : array
        The connectivity of the global NURBS basis.
    cells : array, optional
        If given, use only the given Bezier elements.

    Returns
    -------
    coors : array
        The physical coordinates of the quadrature points of all elements.
    vals : array
        The field variable values in the physical quadrature points.
    dets : array
        The Jacobians of the mapping to the unit reference element in the
        physical quadrature points.
    """
    cdef uint32 ii, ie, n_qp, n_efun, nf, nc, ir, ic
    cdef int32 n_el, n_ep, dim, aux
    cdef uint32 *_cells
    cdef (int32 *) _degrees, _conn, ec
    cdef float64 val
    cdef FMField[1] _bf, _bfg, _det, _vals, _coors
    cdef FMField[1] _bfg_dxi, _dx_dxi, _dxi_dx
    cdef FMField[1] _qp, _variable, _control_points, _weights
    cdef FMField _cs[3]
    cdef FMField[3] _B, _dB_dxi, _N, _dN_dxi
    cdef np.ndarray[float64, mode='c', ndim=2] coors, vals, dets

    if cells is None:
        cells = np.arange(conn.shape[0], dtype=np.uint32)

    n_el = len(cells)
    n_qp = qps.shape[0]
    dim = control_points.shape[1]
    n_efuns = degrees + 1
    n_efun = np.prod(n_efuns)
    nc = variable.shape[1]

    # Output values of the variable.
    vals = np.empty((n_el * n_qp, nc), dtype=np.float64)

    # Output physical coordinates of QPs.
    coors = np.empty((n_el * n_qp, dim), dtype=np.float64)

    # Output Jacobians.
    dets = np.empty((n_el * n_qp, 1), dtype=np.float64)

    # Setup C termporary arrays.
    bf = np.empty((n_efun,), dtype=np.float64)
    bfg = np.empty((dim, n_efun), dtype=np.float64)
    bfg_dxi = np.empty((1, 1, dim, n_efun), dtype=np.float64)
    dx_dxi = np.empty((1, 1, dim, dim), dtype=np.float64)
    dxi_dx = np.empty((1, 1, dim, dim), dtype=np.float64)

    for ii in range(0, <uint32>dim):
        fmf_alloc(_B + ii, n_qp, 1, n_efuns[ii], 1)
        fmf_alloc(_dB_dxi + ii, n_qp, 1, n_efuns[ii], 1)
        fmf_alloc(_N + ii, 1, 1, n_efuns[ii], 1)
        fmf_alloc(_dN_dxi + ii, 1, 1, n_efuns[ii], 1)

    # Assign to C structures.
    array2fmfield1(_bf, bf)
    array2fmfield2(_bfg, bfg)
    array2fmfield4(_bfg_dxi, bfg_dxi)
    array2fmfield4(_dx_dxi, dx_dxi)
    array2fmfield4(_dxi_dx, dxi_dx)
    array2fmfield2(_variable, variable)
    array2fmfield2(_control_points, control_points)
    array2fmfield1(_weights, weights)
    for ii in range(<uint32>dim):
        array2fmfield4(_cs + ii, cs[ii])

    array2pint1(&_degrees, &dim, degrees)
    array2pint2(&_conn, &aux, &n_ep, conn)

    _vals.offset = _coors.offset = _det.offset = _qp.offset = -1
    _vals.nAlloc = _coors.nAlloc = _det.nAlloc = _qp.nAlloc = -1
    _vals.nCell = _coors.nCell = _det.nCell = _qp.nCell = 1
    _vals.nLev = _coors.nLev = _det.nLev = _qp.nLev = 1
    _vals.nRow = _coors.nRow = _det.nRow = _qp.nRow = 1
    _vals.nCol = nc
    _coors.nCol = dim
    _det.nCol = 1
    _qp.nCol = dim

    _vals.val = _vals.val0 = &vals[0, 0]
    _coors.val = _coors.val0 = &coors[0, 0]
    _det.val = _det.val0 = &dets[0, 0]
    _qp.val = _qp.val0 = &qps[0, 0]

    # Pre-compute 1D Bernstein basis B, dB/dxi.
    for iqp in range(0, n_qp):
        for ii in range(0, <uint32>dim):
            nf = n_efuns[ii]
            _eval_bernstein_basis(_B + ii, _dB_dxi + ii,
                                  _qp.val[ii], _degrees[ii])
            (_B + ii).val += nf
            (_dB_dxi + ii).val += nf
        _qp.val += dim

    if is_nurbs(weights):
        # Loop over elements.
        _cells = &cells[0]
        for iseq in range(0, n_el):
            ie = _cells[iseq]

            ec = _conn + n_ep * ie

            _qp.val = _qp.val0
            for ii in range(0, <uint32>dim):
                (_B + ii).val = (_B + ii).val0
                (_dB_dxi + ii).val = (_dB_dxi + ii).val0

            # Loop over quadrature points.
            for iqp in range(0, n_qp):
                _eval_nurbs_basis_tp(_bf, _bfg, _det,
                                     _bfg_dxi,
                                     _dx_dxi, _dxi_dx,
                                     _B, _dB_dxi, _N, _dN_dxi,
                                     _qp, ie,
                                     _control_points, _weights,
                                     _degrees, dim, _cs, _conn, n_el, n_ep,
                                     1, 1)

                # vals[ii, :] = np.dot(bf, variable[ec])
                for ir in range(0, nc):
                    _vals.val[ir] = 0.0

                    for ic in range(0, n_efun):
                        val = _variable.val[ec[ic] * nc + ir]
                        _vals.val[ir] += _bf.val[ic] * val

                # coors[ii, :] = np.dot(bf, control_points[ec])
                for ir in range(0, <uint32>dim):
                    _coors.val[ir] = 0.0

                    for ic in range(0, n_efun):
                        val = _control_points.val[ec[ic] * dim + ir]
                        _coors.val[ir] += _bf.val[ic] * val

                _vals.val += nc
                _coors.val += dim
                _det.val += 1
                _qp.val += dim
                for ii in range(0, <uint32>dim):
                    nf = n_efuns[ii]
                    (_B + ii).val += nf
                    (_dB_dxi + ii).val += nf

    else:
        # Loop over elements.
        _cells = &cells[0]
        for iseq in range(0, n_el):
            ie = _cells[iseq]

            ec = _conn + n_ep * ie

            _qp.val = _qp.val0
            for ii in range(0, <uint32>dim):
                (_B + ii).val = (_B + ii).val0
                (_dB_dxi + ii).val = (_dB_dxi + ii).val0

            # Loop over quadrature points.
            for iqp in range(0, n_qp):
                _eval_bspline_basis_tp(_bf, _bfg, _det,
                                       _bfg_dxi,
                                       _dx_dxi, _dxi_dx,
                                       _B, _dB_dxi, _N, _dN_dxi,
                                       _qp, ie,
                                       _control_points,
                                       _degrees, dim, _cs, _conn, n_el, n_ep,
                                       1, 1)

                # vals[ii, :] = np.dot(bf, variable[ec])
                for ir in range(0, nc):
                    _vals.val[ir] = 0.0

                    for ic in range(0, n_efun):
                        val = _variable.val[ec[ic] * nc + ir]
                        _vals.val[ir] += _bf.val[ic] * val

                # coors[ii, :] = np.dot(bf, control_points[ec])
                for ir in range(0, <uint32>dim):
                    _coors.val[ir] = 0.0

                    for ic in range(0, n_efun):
                        val = _control_points.val[ec[ic] * dim + ir]
                        _coors.val[ir] += _bf.val[ic] * val

                _vals.val += nc
                _coors.val += dim
                _det.val += 1
                _qp.val += dim
                for ii in range(0, <uint32>dim):
                    nf = n_efuns[ii]
                    (_B + ii).val += nf
                    (_dB_dxi + ii).val += nf

    return coors, vals, dets

def eval_in_tp_coors(np.ndarray[float64, mode='c', ndim=2] variable,
                     indices not None,
                     ref_coors not None,
                     np.ndarray[float64, mode='c', ndim=2]
                     control_points not None,
                     np.ndarray[float64, mode='c', ndim=1] weights not None,
                     np.ndarray[int32, mode='c', ndim=1] degrees not None,
                     cs not None,
                     np.ndarray[int32, mode='c', ndim=2] conn not None):
    """
    Evaluate a field variable (if given) or the NURBS geometry in the given
    tensor-product reference coordinates. The field variable is defined by its
    DOFs - the coefficients of the NURBS basis.

    Parameters
    ----------
    variable : array
        The DOF values of the variable with n_c components, shape (:, n_c).
    indices : list of arrays
        The indices of knot spans for each axis, defining the Bezier element
        numbers.
    ref_coors : list of arrays
        The reference coordinates in [0, 1] for each knot span for each axis,
        defining the reference coordinates in the Bezier elements given by
        `indices`.
    control_points : array
        The NURBS control points.
    weights : array
        The NURBS weights.
    degrees : sequence of ints or int
        The basis degrees in each parametric dimension.
    cs : list of lists of 2D arrays
        The element extraction operators in each parametric dimension.
    conn : array
        The connectivity of the global NURBS basis.

    Returns
    -------
    out : array
        The field variable values or NURBS geometry coordinates for the given
        reference coordinates.
    """
    cdef uint32 ii, ip, ie, n_efun, nc, ir, ic, n_vals, uaux
    cdef int32 n_el, n_ep, dim, aux
    cdef (int32 *) _degrees, _conn, ec
    cdef uint32[3] igrid, shape, iis, n_els
    cdef uint32 *_indices[3]
    cdef uint32 **puaux
    cdef FMField[1] _bf, _bfg, _det, _vals, _out
    cdef FMField[1] _bfg_dxi, _dx_dxi, _dxi_dx
    cdef FMField[1] _rc, _control_points, _weights
    cdef FMField[3] _cs, _ref_coors
    cdef np.ndarray[float64, mode='c', ndim=2] _evals, out
    cdef FMField[3] _B, _dB_dxi, _N, _dN_dxi

    dim = control_points.shape[1]
    n_efuns = degrees + 1
    n_efun = np.prod(n_efuns)

    n_vals = 1
    for ii in range(0, <uint32>dim):
        shape[ii] = len(ref_coors[ii])
        n_vals *= shape[ii]

    # Output values.
    if variable is None:
        nc = dim
        out = np.zeros((n_vals, dim), dtype=np.float64)
        _evals = control_points

    else:
        nc = variable.shape[1]
        out = np.zeros((n_vals, nc), dtype=np.float64)
        _evals = variable

    # Setup C termporary arrays.
    rc = np.empty((dim,), dtype=np.float64)
    det = np.empty((1,), dtype=np.float64)
    bf = np.empty((n_efun,), dtype=np.float64)
    bfg = np.empty((dim, n_efun), dtype=np.float64)
    bfg_dxi = np.empty((1, 1, dim, n_efun), dtype=np.float64)
    dx_dxi = np.empty((1, 1, dim, dim), dtype=np.float64)
    dxi_dx = np.empty((1, 1, dim, dim), dtype=np.float64)

    for ii in range(0, <uint32>dim):
        fmf_alloc(_B + ii, 1, 1, n_efuns[ii], 1)
        fmf_alloc(_dB_dxi + ii, 1, 1, n_efuns[ii], 1)
        fmf_alloc(_N + ii, 1, 1, n_efuns[ii], 1)
        fmf_alloc(_dN_dxi + ii, 1, 1, n_efuns[ii], 1)

    # Assign to C structures.
    array2fmfield1(_rc, rc)
    array2fmfield1(_det, det)
    array2fmfield1(_bf, bf)
    array2fmfield2(_bfg, bfg)
    array2fmfield4(_bfg_dxi, bfg_dxi)
    array2fmfield4(_dx_dxi, dx_dxi)
    array2fmfield4(_dxi_dx, dxi_dx)
    array2fmfield2(_control_points, control_points)
    array2fmfield1(_weights, weights)
    array2fmfield2(_vals, _evals)
    for ii in range(<uint32>dim):
        array2fmfield4(_cs + ii, cs[ii])
        n_els[ii] = (_cs + ii).nCell;

    array2pint1(&_degrees, &dim, degrees)
    array2pint2(&_conn, &n_el, &n_ep, conn)

    _out.offset = _out.nAlloc = -1
    _out.nCell = _out.nLev = _out.nRow = 1
    _out.nCol = nc

    _out.val = _out.val0 = &out[0, 0]

    for ii in range(0, <uint32>dim):
        puaux = _indices + ii
        array2puint1(puaux, &uaux, indices[ii])
        array2fmfield1(_ref_coors + ii, ref_coors[ii])

    if is_nurbs(weights):
        for ip in range(0, n_vals):
            _unravel_index(igrid, ip, shape, dim)

            for ii in range(0, <uint32>dim):
                iis[ii] = _indices[ii][igrid[ii]]
                _rc.val[ii] = (_ref_coors + ii).val[igrid[ii]]
            _ravel_multi_index(&ie, iis, n_els, dim)

            _eval_nurbs_basis_tp(_bf, _bfg, _det,
                                 _bfg_dxi,
                                 _dx_dxi, _dxi_dx,
                                 _B, _dB_dxi, _N, _dN_dxi,
                                 _rc, ie,
                                 _control_points, _weights,
                                 _degrees, dim, _cs, _conn, n_el, n_ep,
                                 0, 1)

            # vals[ip, :] = np.dot(bf, variable[ec])
            ec = _conn + n_ep * ie;
            for ir in range(0, nc):
                _out.val[ir] = 0.0

                for ic in range(0, n_efun):
                    _out.val[ir] += _bf.val[ic] * _vals.val[ec[ic] * nc + ir]

            _out.val += nc

    else:
        for ip in range(0, n_vals):
            _unravel_index(igrid, ip, shape, dim)

            for ii in range(0, <uint32>dim):
                iis[ii] = _indices[ii][igrid[ii]]
                _rc.val[ii] = (_ref_coors + ii).val[igrid[ii]]
            _ravel_multi_index(&ie, iis, n_els, dim)

            _eval_bspline_basis_tp(_bf, _bfg, _det,
                                   _bfg_dxi,
                                   _dx_dxi, _dxi_dx,
                                   _B, _dB_dxi, _N, _dN_dxi,
                                   _rc, ie,
                                   _control_points,
                                   _degrees, dim, _cs, _conn, n_el, n_ep,
                                   0, 1)

            # vals[ip, :] = np.dot(bf, variable[ec])
            ec = _conn + n_ep * ie;
            for ir in range(0, nc):
                _out.val[ir] = 0.0

                for ic in range(0, n_efun):
                    _out.val[ir] += _bf.val[ic] * _vals.val[ec[ic] * nc + ir]

            _out.val += nc

    return out



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
Script for plotting 1D DG FEM data stored in VTK files
"""

import glob

import matplotlib
matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt
import argparse

from os.path import join as pjoin
import os
import sys
from glob import glob

from sfepy.discrete.dg.dg_1D_vizualizer import \
    (load_1D_vtks, animate_1D_DG_sol, load_state_1D_vtk, plot1D_legendre_dofs,
    reconstruct_legendre_dofs)


def load_and_plot_fun(folder, filename, t0, t1, tn,
                      ic_fun=None, exact=None,
                      compare=False, polar=False):
    """
    Parameters
    ----------
    folder : str
        folder where to look for files
    filename : str
        used in {name}.i.vtk, i = 0,1, ... tns - 1
    t0 : float
        starting time
    t1 : int
        final time
    tn : int
        number of time steps
    ic_fun : callable
        initital condition
    exact : callable
        exact solution, for transient problems function of space ant time
    compare : bool
    polar : bool
    """

    # load time data
    lmesh, u = load_1D_vtks(folder, filename)
    animate_1D_DG_sol(lmesh, t0, t1, u, tn=tn, ic=ic_fun, exact=exact,
                      delay=100, polar=polar)

    if compare:
        files = glob(pjoin(folder, "*.vtk"))

        n_first = sorted(files)[0].split(".")[-2]
        n_last = sorted(files)[-1].split(".")[-2]

        print("Plotting files {} and {} to compare first and last".format(n_first, n_last))

        coors, u_end = load_state_1D_vtk(pjoin(folder, filename + "." + n_last + ".vtk"))
        coors, u_start = load_state_1D_vtk(pjoin(folder, filename + "." + n_first + ".vtk"))

        plot1D_legendre_dofs(coors, [u_start.swapaxes(0, 1)[:, :, 0], u_end.swapaxes(0, 1)[:, :, 0]])

        plt.figure("Reconstructed solution")
        ww_s, xx = reconstruct_legendre_dofs(coors, None, u_end)
        ww_e, _ = reconstruct_legendre_dofs(coors, None, u_start)
        plt.plot(xx, ww_s[:, 0], label="IC")
        plt.plot(xx, ww_e[:, 0], label="Last")
        plt.legend()
        plt.show()


def main(argv):
    parser = argparse.ArgumentParser(description='Plotting of 1D DG data in VTK files',
                                     epilog='(c) 2019 T. Zitka , Man-machine'
                                           +' Interaction at NTC UWB')
    parser.add_argument("input_name", help="Folder or name of the example in "
                                    + "output folder with VTK data, file names "
                                    + "in format <name>.[0-9]*.vtk. If not "
                                    + "provided asks for the name of the example"
                        , nargs="?")

    parser.add_argument("-s", "--search", help="Search for different orders in "
                                               "provided folder",
                        action="store_true", default=False)
    parser.add_argument("-t0", "--start_time", type=float, default=0,
                        help="Start time of the simulation")
    parser.add_argument("-t1", "--end_time", type=float, default=1.,
                        help="End time of the simulation")
    parser.add_argument("-o", "--order", type=int, default=None,
                        help="Order of the approximation, when example folder"
                             + " cantains more orders this chooses the one")
    parser.add_argument("-cf", "--compare-final", action="store_true",
                        help="To compare starting and final time - " +
                             "usefull for periodic boundary problems")
    parser.add_argument("-p", "--polar", help="Plot in polar projection",
                        action="store_true")

    if argv is None:
        argv = sys.argv[1:]
    args = parser.parse_args(argv)

    t0 = args.start_time
    t1 = args.end_time
    cf = args.compare_final
    pol = args.polar
    if args.input_name is None:
        input_name = str(input("Please provide name of the example in output/" +
                               " folder or full path to data: "))
    else:
        input_name = args.input_name

    if os.path.isdir(input_name):
        full_infolder_path = os.path.abspath(input_name)
    else:
        input_name = pjoin("output", input_name)
        if os.path.isdir(input_name):
            full_infolder_path = os.path.abspath(input_name)
        else:
            print("Example {} not found in {}".format(input_name,
                                                      os.path.abspath("output")))
            return

    print("Input folder found: {}".format(full_infolder_path))
    if args.search:
        print("Input folder contains results for orders {}"
              .format([os.path.basename(fol)
                            for fol in
                                    glob(pjoin(full_infolder_path, "[0-9]*"))]))

        if args.order is None:
            order = str(input("Please provide order of approximation, default is 1: "))
            if len(order) == 0:
                order = 1
            else:
                try:
                    order = int(order)
                except ValueError:
                    print("Value {} for order not understood!".format(order))
                    return
        else:
            order = args.order
        print("Looking for results of order {}".format(order))
        full_infolder_path = pjoin(full_infolder_path, str(order))
        if not os.path.isdir(full_infolder_path):
            print("Input folder with order {} not found".format(full_infolder_path))
            return
    else:
        order = args.order

    contents = glob(pjoin(full_infolder_path, "*.vtk"))
    print()
    tn = len(contents)  # we assume the contents are time step data files
    if tn == 0:
        print("Input folder {} is empty!".format(full_infolder_path))
        return
    base_name = os.path.basename(contents[0]).split(".")[0]
    print("Found {} files, basename is {}".format(tn, base_name))
    print("Plotting ...")
    load_and_plot_fun(full_infolder_path, base_name, t0, t1, tn,
                      compare=cf, polar=pol)


if __name__ == '__main__':
    main(sys.argv[1:])


# -*- Mode: Python -*-
"""
Interface to Lobatto bases.
"""
cimport cython

import numpy as np
cimport numpy as np

cimport sfepy.discrete.common.extmods._fmfield as _f
from sfepy.discrete.common.extmods._fmfield cimport FMField

from sfepy.discrete.common.extmods.types cimport int32, float64, complex128

cdef extern from 'lobatto.h':
    int32 _eval_lobatto1d \
          'eval_lobatto1d'(FMField *out, FMField *coors, int32 order)

    int32 _eval_lobatto_tensor_product \
          'eval_lobatto_tensor_product'(FMField *out, FMField *coors,
                                        int32 *nodes,
                                        float64 cmin, float64 cmax,
                                        int32 diff)

    int32 max_order

@cython.boundscheck(False)
def eval_lobatto1d(np.ndarray[float64, mode='c', ndim=1] coors not None,
                   int32 order):
    """
    Evaluate 1D Lobatto functions of the given order in given points.
    """
    cdef int32 n_coor = coors.shape[0]
    cdef np.ndarray[float64, ndim=1] out = np.zeros(n_coor, dtype=np.float64)
    cdef FMField[1] _coors, _out

    _f.array2fmfield1(_coors, coors)
    _f.array2fmfield1(_out, out)

    _eval_lobatto1d(_out, _coors, order)

    return out

@cython.boundscheck(False)
def eval_lobatto_tensor_product(np.ndarray[float64, mode='c', ndim=2]
                                coors not None,
                                np.ndarray[int32, mode='c', ndim=2]
                                nodes not None,
                                float64 cmin, float64 cmax, int32 order,
                                int32 diff=False):
    """
    Evaluate tensor product Lobatto functions of the given order in given
    points.

    Base functions are addressed using the `nodes` array with rows
    corresponding to individual functions and columns to 1D indices (= orders
    when >= 1) into lobatto[] and d_lobatto[] lists for each axis.
    """
    cdef np.ndarray[float64, ndim=3] out
    cdef FMField[1] _coors, _out
    cdef int32 *_nodes = &nodes[0, 0]
    cdef int32 n_coor = coors.shape[0]
    cdef int32 dim = coors.shape[1]
    cdef int32 n_fun = nodes.shape[0]
    cdef int32 nr

    if (order < 1) or (order > max_order):
        raise ValueError('order must be in [1, %d]! (was %d)'
                         % (max_order, order))

    nr = 1 if not diff else dim
    out = np.zeros((n_coor, nr, n_fun), dtype=np.float64)

    _f.array2fmfield2(_coors, coors)
    _f.array2fmfield3(_out, out)

    _eval_lobatto_tensor_product(_out, _coors, _nodes, cmin, cmax, diff)

    return out



import sys
FREECADPATH = '/usr/lib/freecad/lib/'
sys.path.append(FREECADPATH)

from FreeCAD import Base, newDocument
import Part
import Draft
import ProfileLib.RegularPolygon as Poly
import MeshPart


doc = newDocument()

radius = 0.01
height = 0.1

cyl = doc.addObject("Part::Cylinder", "cyl")
cyl.Radius = radius
cyl.Height = height

sph = doc.addObject("Part::Sphere", "sph")
sph.Radius = radius

uni = doc.addObject("Part::MultiFuse", "uni")
uni.Shapes = [cyl, sph]

ske = doc.addObject('Sketcher::SketchObject', 'Sketch')
ske.Placement = Base.Placement(Base.Vector(0.0, 0.0, 0.0),
                               Base.Rotation(-0.707107, 0.0, 0.0, -0.707107))
Poly.makeRegularPolygon('Sketch', 5,
                        Base.Vector(-1.2 * radius, 0.9 * height, 0),
                        Base.Vector(-0.8 * radius, 0.9 * height, 0))

cut = doc.addObject("PartDesign::Revolution", "Revolution")
cut.Sketch = ske
cut.ReferenceAxis = (ske, ['V_Axis'])
cut.Angle = 360.0

dif = doc.addObject("Part::Cut", "dif")
dif.Base = uni
dif.Tool = cut

cyl1 = doc.addObject("Part::Cylinder", "cyl1")
cyl1.Radius = 0.2 * radius
cyl1.Height = 1.1 * height
cyl1.Placement = Base.Placement(Base.Vector(-1.1 * radius, 0.0, -0.2 * height),
                                Base.Rotation(0.0, 0.0, 0.0, 1))

arr = Draft.makeArray(cyl1, Base.Vector(1, 0, 0), Base.Vector(0, 1, 0), 2, 2)
arr.ArrayType = "polar"
arr.NumberPolar = 6

dif2 = doc.addObject("Part::Cut", "dif2")
dif2.Base = dif
dif2.Tool = arr

cyl2 = doc.addObject("Part::Cylinder", "cyl2")
cyl2.Radius = 0.3 * radius
cyl2.Height = height

dif3 = doc.addObject("Part::Cut", "dif3")
dif3.Base = dif2
dif3.Tool = cyl2

doc.recompute()

Part.export([dif3], 'screwdriver_handle.step')

doc.saveAs('screwdriver_handle.FCStd')

mesh = doc.addObject("Mesh::Feature", "Mesh")
mesh.Mesh = MeshPart.meshFromShape(Shape=dif3.Shape, MaxLength=0.002)
mesh.Mesh.write("./screwdriver_handle.bdf", "NAS", "mesh")



#!/usr/bin/env python
"""SfePy: Simple finite elements in Python

SfePy (simple finite elements in Python) is a software, distributed
under the BSD license, for solving systems of coupled partial
differential equations by the finite element method. The code is based
on NumPy and SciPy packages.
"""

DOCLINES = __doc__.split("\n")

import os
import sys

from build_helpers import (generate_a_pyrex_source, package_check, log,
                           cmdclass, INFO)
# monkey-patch numpy distutils to use Cython instead of Pyrex
from numpy.distutils.command.build_src import build_src

build_src.generate_a_pyrex_source = generate_a_pyrex_source

VERSION = INFO.__version__

CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Programming Language :: C
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: POSIX
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
"""

DOWNLOAD_URL = "http://sfepy.org/doc-devel/downloads.html"

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists('MANIFEST'): os.remove('MANIFEST')

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(ignore_setup_xxx_py=True,
                       assume_default_configuration=True,
                       delegate_options_to_subpackages=True,
                       quiet=True)

    config.add_subpackage('sfepy')

    main_scripts = [
        'phonon.py',
        'extractor.py',
        'homogen.py',
        'postproc.py',
        'probe.py',
        'run_tests.py',
        'simple.py',
        'test_install.py',
    ]

    aux_scripts = [
        'blockgen.py',
        'convert_mesh.py',
        'cylindergen.py',
        'edit_identifiers.py',
        'eval_ns_forms.py',
        'eval_tl_forms.py',
        'extract_edges.py',
        'extract_surface.py',
        'gen_gallery.py',
        'gen_iga_patch.py',
        'gen_lobatto1d_c.py',
        'gen_mesh_prev.py',
        'gen_release_notes.py',
        'gen_solver_table.py',
        'gen_term_table.py',
        'plot_condition_numbers.py',
        'plot_logs.py',
        'plot_mesh.py',
        'plot_quadratures.py',
        'plot_times.py',
        'save_basis.py',
        'show_authors.py',
        'show_mesh_info.py',
        'show_terms_use.py',
        'sync_module_docs.py',
        'tile_periodic_mesh.py',
    ]
    aux_scripts = [os.path.join('script', ii) for ii in aux_scripts]

    config.add_data_files(('sfepy', ('VERSION', 'INSTALL', 'README.rst',
                                     'LICENSE', 'AUTHORS', 'build_helpers.py',
                                     'site_cfg_template.py', 'Makefile')))
    config.add_data_files(('sfepy/script', main_scripts))
    config.add_data_files(('sfepy/script', aux_scripts))

    config.add_data_dir(('sfepy/meshes', 'meshes'))
    config.add_data_dir(('sfepy/examples', 'examples'))
    config.add_data_dir(('sfepy/tests', 'tests'))

    config.get_version('sfepy/version.py')  # sets config.version

    return config

def _mayavi_version(pkg_name):
    try:
        from enthought.mayavi import version

    except:
        from mayavi import version

    return version.version

def _cython_version(pkg_name):
    from Cython.Compiler.Version import version

    return version

def _igakit_version(pkg_name):
    return '0.1'

def _pymetis_version(pkg_name):
    import pymetis

    return pymetis.version

def _scikit_umfpack_version(pkg_name):
    try:
        import scikits.umfpack; scikits.umfpack
        try:
            return scikits.umfpack.__version__

        except AttributeError:
            return '<0.3.1'

    except:
        return None

def check_versions(show_only=False):
    # Cython is a build dependency.
    package_check('cython', INFO.CYTHON_MIN_VERSION,
                  version_getter=_cython_version,
                  show_only=show_only)

    # Check hard and soft dependencies.
    package_check('numpy', INFO.NUMPY_MIN_VERSION,
                  show_only=show_only)
    package_check('scipy', INFO.SCIPY_MIN_VERSION,
                  show_only=show_only)
    package_check('matplotlib', INFO.MATPLOTLIB_MIN_VERSION,
                  show_only=show_only)
    package_check('pyparsing', INFO.PYPARSING_MIN_VERSION,
                  show_only=show_only)
    package_check('tables', INFO.PYTABLES_MIN_VERSION,
                  show_only=show_only)
    package_check(('enthought.mayavi', 'mayavi'),
                  INFO.MAYAVI_MIN_VERSION, optional=True,
                  version_getter=_mayavi_version,
                  show_only=show_only)
    package_check('sympy', INFO.SYMPY_MIN_VERSION, optional=True,
                  messages={'opt suffix' : '; some tests are going to fail!'},
                  show_only=show_only)
    package_check('igakit', INFO.IGAKIT_MIN_VERSION, optional=True,
                  version_getter=_igakit_version,
                  show_only=show_only)
    package_check('petsc4py', INFO.PETSC4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('mpi4py', INFO.MPI4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('slepc4py', INFO.SLEPC4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('pymetis', INFO.PYMETIS_MIN_VERSION, optional=True,
                  version_getter=_pymetis_version,
                  show_only=show_only)
    package_check('scikits.umfpack', INFO.SCIKIT_UMFPACK_MIN_VERSION,
                  optional=True,
                  version_getter=_scikit_umfpack_version,
                  show_only=show_only)
    package_check('meshio', INFO.MESHIO_MIN_VERSION,
                  show_only=show_only)
    package_check('psutil', INFO.PSUTIL_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('pyvista', INFO.PYVISTA_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('opt_einsum', INFO.OPT_EINSUM_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('jax', INFO.JAX_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('dask', INFO.DASK_MIN_VERSION, optional=True,
                  show_only=show_only)

def setup_package():
    if not 'sdist' in sys.argv[1:]:
        # Import setuptools to find a C compiler on windows.
        import setuptools; setuptools

    from numpy.distutils.core import setup

    old_path = os.getcwd()
    local_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    os.chdir(local_path)
    sys.path.insert(0, local_path)
    sys.path.insert(0, os.path.join(local_path, 'sfepy'))  # to retrive version

    # Write the version file.
    fd = open('VERSION', 'w')
    fd.write(VERSION)
    fd.close()

    # Create version.h file.
    filename_in = 'sfepy/discrete/common/extmods/version.h.in'
    filename_out = 'sfepy/discrete/common/extmods/version.h'
    fdi = open(filename_in, 'r')
    fdo = open(filename_out, 'w')
    for line in fdi:
        if line.find('VERSION "0.0.0"') >= 0:
            aux = line.split()
            aux[2] = VERSION
            line = ' '.join(aux) + '\n'
        fdo.write(line)
    fdi.close()
    fdo.close()

    try:
        setup(name='sfepy',
              maintainer="Robert Cimrman",
              maintainer_email="cimrman3@ntc.zcu.cz",
              description=DOCLINES[0],
              long_description="\n".join(DOCLINES[2:]),
              url="http://sfepy.org",
              download_url=DOWNLOAD_URL,
              license='BSD',
              classifiers=list(filter(None, CLASSIFIERS.split('\n'))),
              platforms=["Linux", "Mac OS-X", 'Windows'],
              scripts=['sfepy-run'],
              cmdclass=cmdclass,
              configuration=configuration)

    finally:
        del sys.path[0]
        os.chdir(old_path)

    return

if __name__ == '__main__':
    check_versions()
    setup_package()

    from sfepy import Config
    site_config = Config()
    log.info('\nUsing Python {}.'.format(site_config.python_version()))

    log.info('\nRequired and optional packages found:\n')
    check_versions(show_only=True)



# -*- Mode: Python -*-
cimport cython

@cython.boundscheck(False)
cdef inline int array2fmfield4(FMField *out,
                               np.ndarray[float64, mode='c', ndim=4] arr) \
                               except -1:
    cdef int32 n_cell, n_lev, n_row, n_col
    cdef int32 ii

    sh = arr.shape
    n_cell, n_lev, n_row, n_col = sh[0], sh[1], sh[2], sh[3]

    out.nAlloc = -1
    fmf_pretend(out, n_cell, n_lev, n_row, n_col, &arr[0, 0, 0, 0])

@cython.boundscheck(False)
cdef inline int array2fmfield3(FMField *out,
                               np.ndarray[float64, mode='c', ndim=3] arr) \
                               except -1:
    cdef int32 n_cell = 1, n_lev, n_row, n_col
    cdef int32 ii

    sh = arr.shape
    n_lev, n_row, n_col = sh[0], sh[1], sh[2]

    out.nAlloc = -1
    fmf_pretend(out, n_cell, n_lev, n_row, n_col, &arr[0, 0, 0])

@cython.boundscheck(False)
cdef inline int array2fmfield2(FMField *out,
                               np.ndarray[float64, mode='c', ndim=2] arr) \
                               except -1:
    cdef int32 n_cell = 1, n_lev = 1, n_row, n_col
    cdef int32 ii

    sh = arr.shape
    n_row, n_col = sh[0], sh[1]

    out.nAlloc = -1
    fmf_pretend(out, n_cell, n_lev, n_row, n_col, &arr[0, 0])

@cython.boundscheck(False)
cdef inline int array2fmfield1(FMField *out,
                               np.ndarray[float64, mode='c', ndim=1] arr) \
                               except -1:
    cdef int32 n_cell = 1, n_lev = 1, n_row = 1, n_col
    cdef int32 ii

    sh = arr.shape
    n_col = sh[0]

    out.nAlloc = -1
    fmf_pretend(out, n_cell, n_lev, n_row, n_col, &arr[0])

@cython.boundscheck(False)
cdef inline int array2pint2(int32 **out, int32 *n_row, int32 *n_col,
                            np.ndarray[int32, mode='c', ndim=2] arr) \
                            except -1:
    out[0] = &arr[0, 0]
    n_row[0] = arr.shape[0]
    n_col[0] = arr.shape[1]

@cython.boundscheck(False)
cdef inline int array2pint1(int32 **out, int32 *n_row,
                            np.ndarray[int32, mode='c', ndim=1] arr) \
                            except -1:
    out[0] = &arr[0]
    n_row[0] = arr.shape[0]

@cython.boundscheck(False)
cdef inline int array2puint2(uint32 **out, uint32 *n_row, uint32 *n_col,
                             np.ndarray[uint32, mode='c', ndim=2] arr) \
                             except -1:
    out[0] = &arr[0, 0]
    n_row[0] = arr.shape[0]
    n_col[0] = arr.shape[1]

@cython.boundscheck(False)
cdef inline int array2puint1(uint32 **out, uint32 *n_row,
                             np.ndarray[uint32, mode='c', ndim=1] arr) \
                             except -1:
    out[0] = &arr[0]
    n_row[0] = arr.shape[0]
